@@ -191,7 +191,7 @@ typedef struct _ffecom_symbol_ ffecomSymbol;
 struct _ffecom_symbol_
   {
     tree decl_tree;
-    tree length_tree;		/* For CHARACTER dummies &c. */
+    tree length_tree;		/* For CHARACTER dummies and ASSIGN'ed vars. */
     tree vardesc_tree;		/* For NAMELIST. */
     bool addr;			/* Is address of item instead of item. */
   };
@@ -264,6 +264,8 @@ void ffecom_exec_transition (void);
 void ffecom_expand_let_stmt (ffebld dest, ffebld source);
 #if FFECOM_targetCURRENT == FFECOM_targetGCC
 tree ffecom_expr (ffebld expr);
+tree ffecom_expr_assign (ffebld expr);
+tree ffecom_expr_assign_w (ffebld expr);
 tree ffecom_expr_rw (ffebld expr);
 void ffecom_finish_compile (void);
 void ffecom_finish_decl (tree decl, tree init, bool is_top_level);
@@ -310,6 +312,7 @@ tree ffecom_which_entrypoint_decl (void);
 void emit_nop (void);
 void announce_function (tree decl);
 int count_error (int warningp);
+void error (char *s, ...);
 void expand_decl (tree decl);
 void expand_computed_goto (tree exp);
 void expand_function_end (char *filename, int line, int end_bindings);
