@@ -223,7 +223,6 @@ static char *spaces[]
   "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t      ",	/* 158 */
   "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t       ",	/* 159 */
 };
-static char *filename;
 
 void testname (bool nested, int indent, name first, name last);
 void testnames (bool nested, int indent, int len, name first, name last);
@@ -243,9 +242,9 @@ main (int argc, char **argv)
   struct _name_alpha_ names_alpha;
   name n;
   name newname;
-  char input_name[256];
-  char output_name[256];
-  char include_name[256];
+  char *input_name;
+  char *output_name;
+  char *include_name;
   FILE *incl;
   int fixlengths;
   int total_length;
@@ -262,14 +261,15 @@ main (int argc, char **argv)
   names_alpha.first = (name) &names_alpha;	/* Initialize name order. */
   names_alpha.last = (name) &names_alpha;
 
-  filename = argv[1];
+  if (argc != 4)
+    {
+      fprintf (stderr, "Command form: fini input output-code output-include\n");
+      exit (1);
+    }
 
-  strcpy (input_name, filename);
-  strcpy (input_name + strlen (filename), ".fin");
-  strcpy (output_name, filename);
-  strcpy (output_name + strlen (filename), ".j");
-  strcpy (include_name, filename);
-  strcpy (include_name + strlen (filename), ".h");
+  input_name = argv[1];
+  output_name = argv[2];
+  include_name = argv[3];
 
   in = fopen (input_name, "r");
   if (in == NULL)
