@@ -64,6 +64,7 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 static ffewhereColumnNumber ffelex_final_nontab_column_;	/* 72 or 132. */
 static ffelexType ffelex_first_char_[256];
 static char *ffelex_card_image_;/* Current size is ffelex_max_columns_. */
+static bool ffelex_saw_tab_;	/* True if we saw a tab on the current line. */
 static FILE *ffelex_include_file_;	/* File to start reading from. */
 static ffewhereFile ffelex_include_wherefile_;
 static ffewhereColumnNumber ffelex_max_columns_;	/* Expand these when
@@ -114,7 +115,8 @@ ffewhereColumnNumber cn0, ffewhereLineNumber ln1, ffewhereColumnNumber cn1);
 static void ffelex_display_token_ (void);
 #endif
 static void ffelex_finish_statement_ (void);
-static ffewhereColumnNumber ffelex_handle_tab_ (ffewhereColumnNumber col);
+static ffewhereColumnNumber ffelex_image_char_ (int c,
+						ffewhereColumnNumber col);
 static bool ffelex_is_free_char_ctx_contin_ (ffewhereColumnNumber col);
 static bool ffelex_is_free_nonc_ctx_contin_ (ffewhereColumnNumber col);
 static void ffelex_send_token_ (void);
@@ -169,7 +171,6 @@ ffelex_file_fixed (ffewhereFile wf, FILE *f)
 {
   register int c;		/* Character currently under consideration. */
   register ffewhereColumnNumber column;	/* Not really; 0 means column 1... */
-  register bool saw_tab;	/* True if we saw a tab on the current line. */
   bool disallow_continuation_line;
   bool ignore_disallowed_continuation;
   int last_char_in_file = 0;	/* For getting back into comment-skipping
@@ -264,26 +265,20 @@ beginning_of_line_again:	/* :::::::::::::::::::: */
   /* Skip over comment (and otherwise ignored) lines as quickly as possible! */
 
   while (((lextype = ffelex_first_char_[c]) == FFELEX_typeCOMMENT)
-	 || (lextype == FFELEX_typeERROR) || (lextype == FFELEX_typeSLASH))
+	 || (lextype == FFELEX_typeERROR)
+	 || (lextype == FFELEX_typeSLASH))
     {
       if (lextype == FFELEX_typeERROR)
 	{			/* Bad first character, get line and display
 				   it with message. */
-	  ffelex_card_image_[0] = c;
-	  column = 1;
-	  if (c == '\t')
-	    column = ffelex_handle_tab_ (column);
+	  column = ffelex_image_char_ (c, 0);
 
 	bad_first_character:	/* :::::::::::::::::::: */
 
 	  while (((c = getc (f)) != '\n') && (c != EOF))
 	    {
 	      if (column < FFELEX_columnMAX_ERROR_)
-		{
-		  ffelex_card_image_[column++] = c;
-		  if (c == '\t')
-		    column = ffelex_handle_tab_ (column);
-		}
+		column = ffelex_image_char_ (c, column);
 	    }
 	  ffelex_card_image_[column] = '\0';
 	  ffelex_card_length_ = column;
@@ -299,6 +294,9 @@ beginning_of_line_again:	/* :::::::::::::::::::: */
       else
 	/* Typical case (straight comment), just ignore rest of line. */
 	{
+
+	 comment_line:		/* :::::::::::::::::::: */
+
 	  while ((c != '\n') && (c != EOF))
 	    c = getc (f);
 	}
@@ -320,76 +318,46 @@ beginning_of_line_again:	/* :::::::::::::::::::: */
 	goto end_of_file;	/* :::::::::::::::::::: */
     }				/* while [c, first char, means comment] */
 
+  ffelex_saw_tab_ = FALSE;
+
   if (lextype == FFELEX_typeDEBUG)
     c = ' ';			/* A 'D' or 'd' in column 1 with the
 				   debug-lines option on. */
 
-  /* Non-comment character (like a space or digit).  Read the whole line in
-     very quickly.  Stop after reading the character corresponding to the
-     last column in the card (72 normally, 132 if extend_source is
-     specified). However, if a tab is seen, process it as a typical number of
-     tab stops and let the line be as long as it wants. */
+  column = ffelex_image_char_ (c, 0);
 
-  column = 1;
-  saw_tab = FALSE;
+  /* Read the entire line in as is (with whitespace processing).  */
 
-  ffelex_card_image_[0] = c;
-  if (c == '\t')
+  while (((c = getc (f)) != '\n') && (c != EOF))
     {
-      saw_tab = TRUE;
-      column = ffelex_handle_tab_ (column);
-    }
-
-  /* Read with no checking as long as we've got the space (72 or 132 columns
-     at least).	 This is the typical case, so do it as fast as possible. */
-
-  while (((c = getc (f)) != '\n') && (c != EOF)
-	 && (column < ffelex_final_nontab_column_))
-    {
-      ffelex_card_image_[column++] = c;
-      if (c == '\t')
+      if (column >= ffelex_max_columns_)
 	{
-	  saw_tab = TRUE;
-	  column = ffelex_handle_tab_ (column);
-	}
-    }
+	  ffewhereColumnNumber newmax = ffelex_max_columns_ << 1;
 
-  /* If we saw a tab anywhere so far and until we come to the end of the
-     line, continue reading and allow for growing the card image to
-     accommodate any size line. */
-
-  if (saw_tab)
-    {
-      while ((c != '\n') && (c != EOF))
-	{
-	  if (column >= ffelex_max_columns_)
-	    {
-	      ffelex_card_image_ = malloc_resize_ksr (malloc_pool_image (),
-						      ffelex_card_image_,
-		   (ffelex_max_columns_ << 1) + 1, ffelex_max_columns_ + 1);
-	      ffelex_max_columns_ <<= 1;
-	      if (column >= ffelex_max_columns_)
-		{
-		  ffelex_card_image_[column] = '\0';
-		  ffelex_card_length_ = column;
-		  ffelex_bad_1_ (FFEBAD_LINE_TOO_LONG, ffelex_linecount_,
-				 column + 1);
-		  goto beginning_of_line_again;	/* :::::::::::::::::::: */
-		}
+	  if (newmax <= ffelex_max_columns_)
+	    {			/* Overflowed column number. */
+	      ffelex_card_image_[column] = '\0';
+	      ffelex_card_length_ = column;
+	      ffelex_bad_1_ (FFEBAD_LINE_TOO_LONG, ffelex_linecount_,
+			     column + 1);
+	      goto comment_line;	/* :::::::::::::::::::: */
 	    }
-	  ffelex_card_image_[column++] = c;
-	  if (c == '\t')
-	    column = ffelex_handle_tab_ (column);
-	  c = getc (f);
-	}
-    }
-  else
-    /* Ignore anything past column 72/132. */
-    {
-      while ((c != '\n') && (c != EOF))
-	c = getc (f);
 
-      /* Technically, we should now fill ffelex_card_image_ up to column
+	  ffelex_card_image_
+	    = malloc_resize_ksr (malloc_pool_image (),
+				 ffelex_card_image_,
+				 newmax + 1,
+				 ffelex_max_columns_ + 1);
+	  ffelex_max_columns_ = newmax;
+	}
+      column = ffelex_image_char_ (c, column);
+    }
+
+  /* If no tab, cut off line after column 72/132.  */
+
+  if (!ffelex_saw_tab_ && (column > ffelex_final_nontab_column_))
+    {
+      /* Technically, we should now fill ffelex_card_image_ up thru column
          72/132 with spaces, since character/hollerith constants must count
          them in that manner. To save CPU time in several ways (avoid a loop
          here that would be used only when we actually end a line in
@@ -399,6 +367,7 @@ beginning_of_line_again:	/* :::::::::::::::::::: */
          we encounter end-of-line while actually processing a character
          constant. */
 
+      column = ffelex_final_nontab_column_;
     }
   ffelex_card_image_[column] = '\0';
   ffelex_card_length_ = column;
@@ -545,8 +514,10 @@ stop_looking:			/* :::::::::::::::::::: */
 		       column + 1);
       while ((c = ffelex_card_image_[++column]) == ' ')
 	;
-      if ((c == '\0') || (c == '!') || ((c == '/')
-				&& (ffelex_card_image_[column + 1] == '*')))
+      if ((c == '\0')
+	  || (c == '!')
+	  || ((c == '/')
+	      && (ffelex_card_image_[column + 1] == '*')))
 	{
 	  if (ffelex_strict_ansi_ && (c == '/'))
 	    ffelex_bad_1_ (FFEBAD_NON_ANSI_COMMENT, ffelex_linecount_,
@@ -581,15 +552,18 @@ stop_looking:			/* :::::::::::::::::::: */
 	ffelex_bad_1_ (FFEBAD_NON_ANSI_CONTINUATION_COLUMN, ffelex_linecount_,
 		       column + 1);
       if ((ffelex_raw_mode_ != 0)
-	  && (((c = ffelex_card_image_[column + 1]) != '\0') || !saw_tab))
+	  && (((c = ffelex_card_image_[column + 1]) != '\0')
+	      || !ffelex_saw_tab_))
 	{
 	  ++column;
 	  break;
 	}
       while ((c = ffelex_card_image_[++column]) == ' ')
 	;
-      if ((c == '\0') || (c == '!') || ((c == '/')
-				&& (ffelex_card_image_[column + 1] == '*')))
+      if ((c == '\0')
+	  || (c == '!')
+	  || ((c == '/')
+	      && (ffelex_card_image_[column + 1] == '*')))
 	{
 	  if (ffelex_strict_ansi_ && (c == '/'))
 	    ffelex_bad_1_ (FFEBAD_NON_ANSI_COMMENT, ffelex_linecount_,
@@ -621,15 +595,18 @@ stop_looking:			/* :::::::::::::::::::: */
 	      goto beginning_of_line;	/* :::::::::::::::::::: */
 	    }
 	  if ((ffelex_raw_mode_ != 0)
-	    && (((c = ffelex_card_image_[column + 1]) != '\0') || !saw_tab))
+	      && (((c = ffelex_card_image_[column + 1]) != '\0')
+		  || !ffelex_saw_tab_))
 	    {
 	      ++column;
 	      break;
 	    }
 	  while ((c = ffelex_card_image_[++column]) == ' ')
 	    ;
-	  if ((c == '\0') || (c == '!') || ((c == '/')
-				&& (ffelex_card_image_[column + 1] == '*')))
+	  if ((c == '\0')
+	      || (c == '!')
+	      || ((c == '/')
+		  && (ffelex_card_image_[column + 1] == '*')))
 	    {
 	      if (ffelex_strict_ansi_ && (c == '/'))
 		ffelex_bad_1_ (FFEBAD_NON_ANSI_COMMENT, ffelex_linecount_,
@@ -690,7 +667,7 @@ stop_looking:			/* :::::::::::::::::::: */
 	{
 	  ffewhereColumnNumber i;
 
-	  if (saw_tab || (column >= ffelex_final_nontab_column_))
+	  if (ffelex_saw_tab_ || (column >= ffelex_final_nontab_column_))
 	    {
 	      if (!ffelex_raw_include_)
 		goto beginning_of_line;	/* :::::::::::::::::::: */
@@ -1411,9 +1388,6 @@ ffelex_file_free (ffewhereFile wf, FILE *f)
 {
   register int c;		/* Character currently under consideration. */
   register ffewhereColumnNumber column;	/* Not really; 0 means column 1... */
-  register bool saw_tab;	/* True if we saw a tab on the current line. */
-  register ffewhereColumnNumber coltemp;	/* For lines with initial
-						   blank(s). */
   bool continuation_line;
   ffewhereColumnNumber continuation_column;
   int last_char_in_file;	/* For getting back into comment-skipping
@@ -1503,66 +1477,41 @@ first_line:			/* :::::::::::::::::::: */
 	goto end_of_file;	/* :::::::::::::::::::: */
     }
 
-  /* Non-comment character (like a space or digit).  Read the whole line in
-     very quickly.  Stop after reading the character corresponding to the
-     last column in the card (72 normally, 132 if extend_source is
-     specified). However, if a tab is seen, process it as a typical number of
-     tab stops and let the line be as long as it wants. */
+  ffelex_saw_tab_ = FALSE;
 
-  column = 0;
-  saw_tab = FALSE;
+  column = ffelex_image_char_ (c, 0);
 
-  /* Skip over initial spaces and tabs to see if the first nonblank character
-     is exclamation point, newline, or EOF (line is therefore a comment) or
-     ampersand (line is therefore a continuation line). */
+  /* Read the entire line in as is (with whitespace processing).  */
 
-  while ((c == ' ') || (c == '\t'))
+  while (((c = getc (f)) != '\n') && (c != EOF))
     {
-      column++;
-      if (c == '\t')
+      if (column >= ffelex_max_columns_)
 	{
-	  saw_tab = TRUE;
-	  column = ffelex_handle_tab_ (column);
+	  ffewhereColumnNumber newmax = ffelex_max_columns_ << 1;
+
+	  if (newmax <= ffelex_max_columns_)
+	    {			/* Overflowed column number. */
+	      ffelex_card_image_[column] = '\0';
+	      ffelex_card_length_ = column;
+	      ffelex_bad_1_ (FFEBAD_LINE_TOO_LONG, ffelex_linecount_,
+			     column + 1);
+	      goto comment_line;	/* :::::::::::::::::::: */
+	    }
+
+	  ffelex_card_image_
+	    = malloc_resize_ksr (malloc_pool_image (),
+				 ffelex_card_image_,
+				 newmax + 1,
+				 ffelex_max_columns_ + 1);
+	  ffelex_max_columns_ = newmax;
 	}
-      c = getc (f);
+      column = ffelex_image_char_ (c, column);
     }
 
-  continuation_column = 0;
+  /* If no tab, cut off line after column 132.  */
 
-  switch (c)
-    {
-    case '!':
-    case '\n':
-    case EOF:
-      goto comment_line;	/* :::::::::::::::::::: */
-
-    case '&':
-      continuation_column = column + 1;
-      /* Fall through. */
-    default:
-      for (coltemp = 0; coltemp < column; coltemp++)
-	ffelex_card_image_[coltemp] = ' ';
-    }
-
-  /* Read with no checking as long as we've got the space (132 columns). This
-     is the typical case, so do it as fast as possible. */
-
-  while ((c != '\n') && (c != EOF)
-	 && (column < FFELEX_FREE_MAX_COLUMNS_))
-    {
-      ffelex_card_image_[column++] = c;
-      if (c == '\t')
-	{
-	  saw_tab = TRUE;
-	  column = ffelex_handle_tab_ (column);
-	}
-      c = getc (f);
-    }
-
-  /* Ignore anything past column FFELEX_FREE_MAX_COLUMNS_. */
-
-  while ((c != '\n') && (c != EOF))
-    c = getc (f);
+  if (!ffelex_saw_tab_ && (column > FFELEX_FREE_MAX_COLUMNS_))
+    column = FFELEX_FREE_MAX_COLUMNS_;
 
   ffelex_card_image_[column] = '\0';
   ffelex_card_length_ = column;
@@ -1572,6 +1521,29 @@ first_line:			/* :::::::::::::::::::: */
 
   last_char_in_file = c;	/* Should be either '\n' or EOF. */
 
+  column = 0;
+  continuation_column = 0;
+
+  /* Skip over initial spaces to see if the first nonblank character
+     is exclamation point, newline, or EOF (line is therefore a comment) or
+     ampersand (line is therefore a continuation line). */
+
+  while ((c = ffelex_card_image_[column]) == ' ')
+    ++column;
+
+  switch (c)
+    {
+    case '!':
+    case '\0':
+      goto beginning_of_line;	/* :::::::::::::::::::: */
+
+    case '&':
+      continuation_column = column + 1;
+      /* Fall through. */
+    default:
+      break;
+    }
+
   /* The line definitely has content of some kind, install new end-statement
      point for error messages. */
 
@@ -1584,32 +1556,33 @@ first_line:			/* :::::::::::::::::::: */
 
   if (continuation_line)
     {
-      column = continuation_column;
       if (continuation_column == 0)
 	{
 	  if (ffelex_raw_mode_ != 0)
 	    {
 	      ffelex_bad_1_ (FFEBAD_BAD_CHAR_CONTINUE, ffelex_linecount_,
-			     coltemp + 1);
+			     column + 1);
 	    }
 	  else if (ffelex_token_->type != FFELEX_typeNONE)
 	    {
 	      ffelex_bad_1_ (FFEBAD_BAD_LEXTOK_CONTINUE, ffelex_linecount_,
-			     coltemp + 1);
+			     column + 1);
 	    }
 	}
-      else if (ffelex_is_free_char_ctx_contin_ (column))
+      else if (ffelex_is_free_char_ctx_contin_ (continuation_column))
 	{			/* Line contains only a single "&" as only
 				   nonblank character. */
-	  ffelex_bad_1_ (FFEBAD_BAD_FREE_CONTINUE, ffelex_linecount_, column);
+	  ffelex_bad_1_ (FFEBAD_BAD_FREE_CONTINUE, ffelex_linecount_,
+			 continuation_column);
 	  goto beginning_of_line;	/* :::::::::::::::::::: */
 	}
+      column = continuation_column;
     }
   else
     column = 0;
 
-  continuation_line = FALSE;
   c = ffelex_card_image_[column];
+  continuation_line = FALSE;
 
   /* Here is the main engine for parsing.  c holds the character at column.
      It is already known that c is not a blank, end of line, or shriek,
@@ -2380,7 +2353,8 @@ ffelex_init_1 ()
 
   ffelex_final_nontab_column_ = 72;
   ffelex_max_columns_ = FFELEX_columnINITIAL_SIZE_;
-  ffelex_card_image_ = malloc_new_ksr (malloc_pool_image (), "FFELEX card image",
+  ffelex_card_image_ = malloc_new_ksr (malloc_pool_image (),
+				       "FFELEX card image",
 				       FFELEX_columnINITIAL_SIZE_ + 1);
   ffelex_card_image_[0] = '\0';
 
@@ -2391,7 +2365,7 @@ ffelex_init_1 ()
   ffelex_first_char_['\n'] = FFELEX_typeCOMMENT;
   ffelex_first_char_['\v'] = FFELEX_typeCOMMENT;
   ffelex_first_char_['\f'] = FFELEX_typeCOMMENT;
-  ffelex_first_char_['\r'] = FFELEX_typeCOMMENT;
+  ffelex_first_char_['\r'] = FFELEX_typeRAW;
   ffelex_first_char_[' '] = FFELEX_typeRAW;
   ffelex_first_char_['!'] = FFELEX_typeCOMMENT;
   ffelex_first_char_['*'] = FFELEX_typeCOMMENT;
@@ -3359,21 +3333,27 @@ ffelex_finish_statement_ ()
     }
 }
 
-/* ffelex_handle_tab_ -- Deal with tab in ffelex_card_image_, adjust column number
+/* "Image" a character onto the card image, return incremented column number.
 
-   ffewhereColumnNumber c;
-   c = ffelex_handle_tab_(c);
+   Normally invoking this function as in
+     column = ffelex_image_char_ (c, column);
+   is the same as doing:
+     ffelex_card_image_[column++] = c;
 
-   Assumes a tab is in ffelex_card_image_[c - 1], replaces it with a space and
-   inserts an appropriate number of subsequent spaces, returning the new
-   column value.  "Appropriate" is to the next tab position, where tab
-   positions start in column 9 and each eighth column afterwards.
+   However, tabs and carriage returns are handled specially, to preserve
+   the visual "image" of the input line (in most editors) in the card
+   image.
 
-   Overwrite tab with space in ffelex_card_image_.
-   Calculate how many additional spaces should be written.
-   Write that many spaces.
-   Add that number to the column value.
-   Return the column value.
+   Carriage returns are ignored, as they are assumed to be followed
+   by newlines.
+
+   A tab is handled by first doing:
+     ffelex_card_image_[column++] = ' ';
+   That is, it translates to at least one space.  Then, as many spaces
+   are imaged as necessary to bring the column number to the next tab
+   position, where tab positions start in the ninth column and each
+   eighth column afterwards.  ALSO, a static var named ffelex_saw_tab_
+   is set to TRUE to notify the lexer that a tab was seen.
 
    Columns are numbered and tab stops set as illustrated below:
 
@@ -3382,32 +3362,29 @@ ffelex_finish_statement_ ()
    xx	   yy	   zz
    ...
    xxxxxxx yyyyyyy zzzzzzz
-   xxxxxxxx	   yyyyyyyy...
-
-   When this function is called, the column number has already been
-   incremented by 1, so it is never zero.  We calculate how many more
-   spaces are needed, a value with range 0-7 according to the following
-   column status:
-
-   c % 8: 01234567
-   ++:	  07654321  */
+   xxxxxxxx	   yyyyyyyy...  */
 
 static ffewhereColumnNumber
-ffelex_handle_tab_ (ffewhereColumnNumber c)
+ffelex_image_char_ (int c, ffewhereColumnNumber column)
 {
-  ffewhereColumnNumber spaces;
+  switch (c)
+    {
+    case '\r':
+      break;
 
-  assert (c != 0);
+    case '\t':
+      ffelex_card_image_[column++] = ' ';
+      while ((column & 7) != 0)
+	ffelex_card_image_[column++] = ' ';
+      break;
 
-  ffelex_card_image_[c - 1] = ' ';
-  spaces = (8 - (c % 8)) % 8;
+    default:
+      ffelex_saw_tab_ = TRUE;
+      ffelex_card_image_[column++] = c;
+      break;
+    }
 
-  assert (spaces < 8);
-
-  while (spaces-- > 0)
-    ffelex_card_image_[c++] = ' ';
-
-  return c;
+  return column;
 }
 
 /* ffelex_is_free_char_ctx_contin_ -- Character Context Continuation?
