@@ -78,6 +78,9 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 #ifndef FFETARGET_defaultIS_UGLY
 #define FFETARGET_defaultIS_UGLY 0
 #endif
+#ifndef FFETARGET_defaultIS_UGLY_ARGS
+#define FFETARGET_defaultIS_UGLY_ARGS 0
+#endif
 #ifndef FFETARGET_defaultIS_VXT_NOT_90
 #define FFETARGET_defaultIS_VXT_NOT_90 0
 #endif
@@ -885,7 +888,8 @@ void *ffetarget_memcpy_ (void *dst, void *src, size_t len);
 #ifdef REAL_ARITHMETIC
 #define ffetarget_convert_complex1_integer(res,l) \
   ({ REAL_VALUE_TYPE resi, resr; \
-     REAL_VALUE_FROM_INT (resr, (long) (l), (long) 0); \
+     ffetargetInteger1 lf = (l); \
+     REAL_VALUE_FROM_INT (resr, (long) lf, (long) ((lf < 0) ? -1 : 0)); \
      resi = dconst0; \
      REAL_VALUE_TO_TARGET_SINGLE (resr, (res)->real); \
      REAL_VALUE_TO_TARGET_SINGLE (resi, (res)->imaginary); \
@@ -930,11 +934,12 @@ void *ffetarget_memcpy_ (void *dst, void *src, size_t len);
 #endif
 #ifdef REAL_ARITHMETIC
 #define ffetarget_convert_complex2_integer(res,l) \
-  ({REAL_VALUE_TYPE resi, resr; \
-    REAL_VALUE_FROM_INT (resr, (long) (l), (long) 0); \
-    resi = dconst0; \
-    REAL_VALUE_TO_TARGET_DOUBLE (resr, (long *) &((res)->real.v[0])); \
-    REAL_VALUE_TO_TARGET_DOUBLE (resi, (long *) &((res)->imaginary.v[0])); \
+  ({ REAL_VALUE_TYPE resi, resr; \
+     ffetargetInteger1 lf = (l); \
+     REAL_VALUE_FROM_INT (resr, (long) lf, (long) ((lf < 0) ? -1 : 0)); \
+     resi = dconst0; \
+     REAL_VALUE_TO_TARGET_DOUBLE (resr, (long *) &((res)->real.v[0])); \
+     REAL_VALUE_TO_TARGET_DOUBLE (resi, (long *) &((res)->imaginary.v[0])); \
      FFEBAD; })
 #else
 #define ffetarget_convert_complex2_integer(res,l) \
@@ -1035,7 +1040,8 @@ void *ffetarget_memcpy_ (void *dst, void *src, size_t len);
 #ifdef REAL_ARITHMETIC
 #define ffetarget_convert_real1_integer1(res,l) \
   ({ REAL_VALUE_TYPE resr; \
-     REAL_VALUE_FROM_INT (resr, (long) (l), (long) 0); \
+     ffetargetInteger1 lf = (l); \
+     REAL_VALUE_FROM_INT (resr, (long) lf, (long) ((lf < 0) ? -1 : 0)); \
      REAL_VALUE_TO_TARGET_SINGLE (resr, *(res)); \
      FFEBAD; })
 #else
@@ -1060,7 +1066,8 @@ void *ffetarget_memcpy_ (void *dst, void *src, size_t len);
 #ifdef REAL_ARITHMETIC
 #define ffetarget_convert_real2_integer(res,l) \
   ({ REAL_VALUE_TYPE resr; \
-     REAL_VALUE_FROM_INT (resr, (long) (l), (long) 0); \
+     ffetargetInteger1 lf = (l); \
+     REAL_VALUE_FROM_INT (resr, (long) lf, (long) ((lf < 0) ? -1 : 0)); \
      REAL_VALUE_TO_TARGET_DOUBLE (resr, (long *) &((res)->v[0])); \
      FFEBAD; })
 #define ffetarget_convert_real2_integer1 ffetarget_convert_real2_integer
