@@ -68,6 +68,7 @@ bool ffe_is_mainprog_;		/* TRUE if current prog unit known to be
 bool ffe_is_pedantic_ = FFETARGET_defaultIS_PEDANTIC;
 bool ffe_is_saveall_;		/* TRUE if mainprog or SAVE (no args) seen. */
 bool ffe_is_ugly_ = FFETARGET_defaultIS_UGLY;
+bool ffe_is_ugly_args_ = FFETARGET_defaultIS_UGLY_ARGS;
 bool ffe_is_version_ = FALSE;
 bool ffe_is_vxt_not_90_ = FFETARGET_defaultIS_VXT_NOT_90;
 ffeCase ffe_case_intrin_ = FFETARGET_defaultCASE_INTRIN;
@@ -163,6 +164,10 @@ ffe_decode_option (char *opt)
 	ffe_set_is_ugly (TRUE);
       else if (strcmp (&opt[2], "no-ugly") == 0)
 	ffe_set_is_ugly (FALSE);
+      else if (strcmp (&opt[2], "ugly-args") == 0)
+	ffe_set_is_ugly_args (TRUE);
+      else if (strcmp (&opt[2], "no-ugly-args") == 0)
+	ffe_set_is_ugly_args (FALSE);
       else if (strcmp (&opt[2], "debug") == 0)
 	ffe_set_is_ffedebug (TRUE);
       else if (strcmp (&opt[2], "no-debug") == 0)
